# unfitted Heat equation with Neumann b.c.
# higher order geometry approximation with space-time mesh transformation
from ngsolve import *
from time import sleep
from netgen.geom2d import unit_square
from netgen.geom2d import SplineGeometry
from netgen.meshing import MeshingParameters

from ngsolve.internal import *
from ngsolve.solvers import *
from ngsolve.meshes import *
from xfem import *
from math import pi
import sys

from xfem.lset_spacetime import *

ngsglobals.msg_level = 1

i = 2
gamma = 0.05

if hasattr(sys, 'argv') and len(sys.argv) == 5 and sys.argv[1] == "i" and sys.argv[3] == "stab":
    i = int(sys.argv[2])
    gamma = float(sys.argv[4])
    print("Sys argv :", sys.argv)
    print("Loading manual val for i, gamma: ", i, gamma)
else:
    print("Loading default val for i, gamma: ", i, gamma)

Patchint_sliced = False # Only for p4 right now...

square = SplineGeometry()
square.AddRectangle([-0.6,-1],[0.6,1])
ngmesh = square.GenerateMesh(maxh=0.5**(i+1), quad_dominated=False)
mesh = Mesh (ngmesh)

#x_len = 0.6
#y_len = 1.2
#mesh = MakeStructured2DMesh(quads=False,nx=2**(i+1),ny=2**(i+2),mapping= lambda x,y : (2*x_len*x-x_len,2*y_len*y-y_len))

# polynomial order in time
k_t = 3
# polynomial order in space
k_s = k_t
# spatial FESpace for solution
fes1 = H1(mesh, order=k_s)
# polynomial order in time for level set approximation
lset_order_time = k_t
# integration order in time
time_order = 2*k_t
# time finite element (nodal!)
tfe = ScalarTimeFE(k_t) 
# space-time finite element space
st_fes = SpaceTimeFESpace(fes1,tfe, flags = {"dgjumps": True})

#Fitted heat equation example
tstart = 0
tend = 0.5
delta_t = (tend - tstart)/(2**(i))
tnew = 0

told = Parameter(tstart)
tref = ReferenceTimeVariable()
t = told + delta_t*tref

lset_adap_st = LevelSetMeshAdaptation_Spacetime(mesh, order_space = k_s, order_time = lset_order_time,
                                                threshold=0.5, discontinuous_qn=True)

# radius of disk (the geometry)
r0 = 0.5
r1 = 0.5*pi/r0                                        

# position shift of the geometry in time
rho =  CoefficientFunction((1/(pi))*sin(2*pi*t))
rhoL = lambda t:CoefficientFunction((1/(pi))*sin(2*pi*t))
# velocity of position shift
d_rho = CoefficientFunction(2*cos(2*pi*t))

# rho =  CoefficientFunction(1*t)
# rhoL = lambda t:CoefficientFunction(1*t)
# # velocity of position shift
# d_rho = CoefficientFunction(1)

# rho =  CoefficientFunction(0)
# rhoL = lambda t:CoefficientFunction(0)
# # velocity of position shift
# d_rho = CoefficientFunction(1)

alpha = 1
#convection velocity:
w = CoefficientFunction((0,d_rho)) 
# w = 10*CoefficientFunction((-y,x)) 

# level set
r = sqrt(x**2+(y-rho)**2)
rL = lambda t : sqrt(x**2+(y-rhoL(t))**2)
# levelset= r - r0

levelset= CoefficientFunction(r -r0)

# solution and r.h.s.
Q = pi/r0

# coeff_f = CoefficientFunction(  -(pi/r0)*r1*( sin(r1*r)**2 - cos(r1*r)**2 )  + (pi/r0)*cos(r1*r)*sin(r1*r)*(1/r) )
# u_exact = CoefficientFunction(cos(r1*r)*cos(r1*r) ) - 1
# u_exactL = lambda t: CoefficientFunction(cos(r1*rL(t))*cos(r1*rL(t))) - 1

coeff_f = (Q/r * sin(Q*r) + (Q**2) * cos(Q*r)) * sin(pi*t) + pi * cos(Q*r) * cos(pi*t)
u_exact = cos(Q*r) * sin(pi*t)
u_exactL = lambda t: cos(Q*sqrt(x**2+(y-rhoL(t))**2)) * sin(pi*t)

# coeff_f = (Q/r * sin(Q*r) + (Q**2) * cos(Q*r))
# u_exact = cos(Q*r) + 1
# u_exactL = lambda t: cos(Q*sqrt(x**2+(y-rhoL(t))**2)) + 1

# coeff_f = 0
# u_exact = 1
# u_exactL = lambda t: 1


gfu = GridFunction(st_fes)

u_ic = CreateTimeRestrictedGF(gfu,0)
u_last = CreateTimeRestrictedGF(gfu,1.0)
u_first = CreateTimeRestrictedGF(gfu,1.0)

u,v = st_fes.TnT()


lset_p1 = lset_adap_st.lset_p1    


lset_top = CreateTimeRestrictedGF(lset_p1,1.0)
lset_bottom = CreateTimeRestrictedGF(lset_p1,0.0)

dfm = lset_adap_st.deform
dfm_last_top = CreateTimeRestrictedGF(dfm,1.0)
dfm_current_top = CreateTimeRestrictedGF(dfm,1.0)
dfm_current_bottom = CreateTimeRestrictedGF(dfm,0.0)
dfm_time_node = CreateTimeRestrictedGF(dfm,0.)

t_old = tstart

h = specialcf.mesh_size

#Draw(lset_top,mesh,"lset")
#Draw(IfPos(-lset_top,u_exactL(told),float('nan')),mesh,"u_exact")
#Draw(IfPos(-lset_top,u_ic,float('nan')),mesh,"u")

#Draw(dfm,mesh,"dfm")
#Draw(dfm_current_bottom,mesh,"dfm_current_bottom")
#Draw(dfm_current_top,mesh,"dfm_current_top")

#visoptions.deformation = 1

# Draw(u_last-u_exactL(told+delta_t),mesh,"ulast")
Draw(IfPos(lset_top,0,u_last),mesh,"ulast")

#Draw(h, mesh, "h")

lset_neg = { "levelset" : lset_p1, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_bottom = { "levelset" : lset_bottom, "domain_type" : NEG, "subdivlvl" : 0}
lset_neg_top = { "levelset" : lset_top, "domain_type" : NEG, "subdivlvl" : 0}

def SpaceTimeNegBFI(form):
    return SymbolicBFI(levelset_domain = lset_neg, form = form, time_order=time_order, definedonelements = ci.GetElementsOfType(HASNEG))

ci = CutInfo(mesh,time_order=time_order)

    
hasneg_integrators_a = []
hasneg_integrators_a_top = []
hasneg_integrators_f = []
hasneg_integrators_f_bottom = []
patch_integrators_a = []

# hasneg_integrators_a.append(SpaceTimeNegBFI(form =  -u*(dt(v) + InnerProduct( dt_vec(dfm) , grad(v)) )))
hasneg_integrators_a.append(SpaceTimeNegBFI(form =  v*(dt(u) + InnerProduct( dt_vec(dfm) , grad(u)) )))
hasneg_integrators_a.append(SpaceTimeNegBFI(form = alpha * delta_t* InnerProduct( grad(u), grad(v)) ))
hasneg_integrators_a.append(SpaceTimeNegBFI(form = delta_t*v*InnerProduct(w,grad(u))))
#hasneg_integrators_a.append(SpaceTimeNegBFI(form = u*v))
#hasneg_integrators_a.append(SymbolicBFI(levelset_domain = lset_neg_top, form = fix_t(u,1)*fix_t(v,1), deformation = dfm_current_top))
hasneg_integrators_a_top.append(SymbolicBFI(levelset_domain = lset_neg_bottom, form = fix_t(u,0)*fix_t(v,0)))

patch_integrators_a.append(SymbolicFacetPatchBFI(form = h**(-2)*delta_t*(1+delta_t/h)*gamma*(u - u.Other() )*(v - v.Other()), skeleton=False, time_order=time_order))

timerule = [(0.0469101, 0.118463), (0.230765, 0.239314), (0.5, 0.284444), (0.769235, 0.239314), (0.95309, 0.118463) ]

def getFacetPatchatTime(tval):
    return SymbolicFacetPatchBFI(form = h**(-2)*delta_t*(1+delta_t/h)*gamma*(fix_t(u,tval) - fix_t(u.Other(), tval) )*( fix_t(v,tval) - fix_t( v.Other(), tval)), skeleton=False)

hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg, form = delta_t*coeff_f*v, time_order=time_order))
#hasneg_integrators_f.append(SymbolicLFI(levelset_domain = lset_neg_bottom, form = u_ic*fix_t(v,0), deformation = dfm_current_bottom))
hasneg_integrators_f_bottom.append(SymbolicLFI(levelset_domain = lset_neg_bottom,form = u_ic*fix_t(v,0)))

outfile = open("errorp"+str(k_t)+"_dev_i"+str(i)+"_gamma"+str(gamma)+".dat", "w")

l2max = 0
l2_timeintmax = 0
maxdists = []
while tend - t_old > delta_t/2:
    # update lset geometry to new time slab (also changes lset_p1 !)
    #dfm = lset_adap_st.CalcDeformation(levelset,told,t_old,delta_t)
    #with TaskManager():
    dfm = lset_adap_st.CalcDeformation(levelset)
    mesh.SetDeformation(dfm)
    maxdist = lset_adap_st.CalcMaxDistance(levelset)
    mesh.UnsetDeformation()
    #print("maxdist=",maxdist) #, given_pts = [])
    maxdists.append(maxdist)
    
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=0.0,space_gf=lset_bottom)
    RestrictGFInTime(spacetime_gf=lset_p1,reference_time=1.0,space_gf=lset_top)
    RestrictGFInTime(spacetime_gf=dfm,reference_time=0.0,space_gf=dfm_current_bottom)   
    RestrictGFInTime(spacetime_gf=dfm,reference_time=1.0,space_gf=dfm_current_top)   

    if t_old == tstart:
    #if True:
        mesh.SetDeformation(dfm_current_bottom)
        u_ic.Set(CoefficientFunction(u_exactL(tstart)))
        #u_ic.Set(CoefficientFunction(u_exactL(t_old)))
        mesh.UnsetDeformation()
    else:
        u_ic.Set(shifted_eval(u_last, back = dfm_last_top, forth = dfm_current_bottom))


    # update markers in (space-time) mesh
    ci.Update(lset_p1,time_order=time_order)

    # re-compute the facets for stabilization:
    ba_facets = GetFacetsWithNeighborTypes(mesh,a=ci.GetElementsOfType(HASNEG),
                                                b=ci.GetElementsOfType(IF))
    
    # Draw(BitArrayCF(ci.GetElementsOfType(NEG)),mesh,"NEG")
    # Draw(BitArrayCF(ci.GetElementsOfType(IF)),mesh,"IF")
    # Draw(BitArrayCF(GetElementsWithNeighborFacets(mesh,ba_facets)),mesh,"stabilized")
    # input("")
    # re-evaluate the "active dofs" in the space time slab
    active_dofs = GetDofsOfElements(st_fes,ci.GetElementsOfType(HASNEG))

    # re-set definedonelements-markers according to new markings:
    for integrator in hasneg_integrators_a +hasneg_integrators_a_top + hasneg_integrators_f + hasneg_integrators_f_bottom:
        integrator.SetDefinedOnElements(ci.GetElementsOfType(HASNEG))
    for integrator in patch_integrators_a:
        integrator.SetDefinedOnElements(ba_facets)

    #a = BilinearForm(st_fes,check_unused=False,symmetric=False)
    a = RestrictedBilinearForm(st_fes,"a", ci.GetElementsOfType(HASNEG), ba_facets, check_unused=False)
    for integrator in hasneg_integrators_a: # + patch_integrators_a:
        a += integrator
    if not Patchint_sliced:
        for integrator in patch_integrators_a:
            a += integrator
    
    a_top = RestrictedBilinearForm(st_fes,"a_top", ci.GetElementsOfType(HASNEG), ba_facets, check_unused=False)
    for integrator in hasneg_integrators_a_top:
        a_top += integrator
    
    f = LinearForm(st_fes)
    for integrator in hasneg_integrators_f:
        f += integrator

    f_bottom = LinearForm(st_fes)
    for integrator in hasneg_integrators_f_bottom:
        f_bottom += integrator
    
    
    # update mesh deformation and assemble linear system
    mesh.SetDeformation(dfm)
    #with TaskManager():
    a.Assemble()
    f.Assemble()
    mesh.UnsetDeformation()
    
    mesh.SetDeformation(dfm_current_bottom)
    f_bottom.Assemble()
    mesh.UnsetDeformation()
    
    mesh.SetDeformation(dfm_current_bottom)
    a_top.Assemble()
    mesh.UnsetDeformation()
    
    a.mat.AsVector().data = a.mat.AsVector() + a_top.mat.AsVector() #+ a_no_dfm.mat.AsVector()
    f.vec.data = f.vec + f_bottom.vec
    
    if Patchint_sliced:
        for (tval, weight) in timerule:
            print("tval, weight = ", tval, weight)
            
            a_patch_time_node = RestrictedBilinearForm(st_fes,"a_no_dfm", ci.GetElementsOfType(HASNEG), ba_facets, check_unused=False)
            integrator = getFacetPatchatTime(tval)
            integrator.SetDefinedOnElements(ba_facets)
            a_patch_time_node += integrator
            
            RestrictGFInTime(spacetime_gf=dfm,reference_time=tval,space_gf=dfm_time_node)
            
            mesh.SetDeformation(dfm_time_node)
            a_patch_time_node.Assemble()
            mesh.UnsetDeformation()
            
            a.mat.AsVector().data = a.mat.AsVector() + weight* a_patch_time_node.mat.AsVector()
    
    # solve linear system
    #pre = a.mat.Inverse(active_dofs,"")
    #inv = CGSolver(a.mat, pre, printrates=True, precision=1e-10, maxsteps=15)
    #inv = GMRESSolver(mat=a.mat,pre=pre,printrates=True,precision=1e-8,maxsteps=15)
    pre = a.mat.Inverse(active_dofs,"")
    inv = CGSolver(a.mat,pre,printrates=True,maxsteps=5)
    # inv = a.mat.Inverse(active_dofs,inverse="pardiso")
    gfu.vec.data = inv* f.vec.data
    #gfu.vec.data = GMRes( a.mat, f.vec, pre, tol=1e-20, printrates=True)
    
    # tmp = f.vec.CreateVector()
    # for it in range(5):
    #     tmp.data = f.vec.data - a.mat * gfu.vec.data
    #     print("Norm of tmp: ", Norm(tmp))
    #     gfu.vec.data = gfu.vec.data + inv * tmp.data
    
    # evaluate upper trace of solution for
    #  * for error evaluation 
    #  * upwind-coupling to next time slab
    RestrictGFInTime(spacetime_gf=gfu,reference_time=1.0,space_gf=u_last)
    
    # compute error at final time
    mesh.SetDeformation(dfm_current_top)
    l2error = sqrt(Integrate({ "levelset" : lset_top, "domain_type" : NEG}, (u_exactL(t_old + delta_t) - u_last)**2, mesh, order=2*k_s))

    mass = Integrate({ "levelset" : lset_top, "domain_type" : NEG}, u_last, mesh, order=2*k_s)
    # print("mass=",mass)
    mesh.UnsetDeformation()
    
    mesh.SetDeformation(dfm)
    l2error_time_int = sqrt(Integrate({ "levelset" : lset_p1, "domain_type" : NEG}, (u_exactL(t) - gfu)**2, mesh, order=2*k_s, time_order=time_order))
    mesh.UnsetDeformation()
    
    # update time variable (float and ParameterCL)
    t_old = t_old + delta_t
    told.Set(t_old)
    
    #Draw(lset_top, mesh, "lset_top")
    #Draw(lset_bottom, mesh, "lset_bottom")
    #Draw(IfPos(-lset_top,(u_exactL(told)-u_last)**2,float('nan')),mesh,"error")
    #nf = IndicatorCF( mesh=mesh, ba=ci.GetElementsOfType(HASNEG))
    #Draw( nf, mesh, "NEG")
    
    
    # print time and error
    #print("\rt = {0:10}, l2error = {1:20}".format(t_old,l2error),end="")
    # print("t = ",t_old, " l2error = ",l2error, " time int L2: ", l2error_time_int)
    print("\rt = ",t_old,"            ",end="")
    #outfile.write(str(t_old)+"\t"+str(l2error)+"\t"+str(Norm(tmp))+"\n")
    outfile.write(str(t_old)+"\t"+str(l2error)+"\t"+str(l2error_time_int)+"\n")
    
    #least_abs_lset_val = 10.
    #for v in lset_p1.vec:
        #if abs(v) < least_abs_lset_val:
            #least_abs_lset_val = abs(v)
    
    #print("least abs: \t\t\t\t\t\t", least_abs_lset_val)
    
    if l2error > l2max:
        l2max = l2error
    if l2error_time_int > l2_timeintmax:
        l2_timeintmax = l2error_time_int
    
    # Redraw:
    #Redraw(blocking=True)

    # store deformation at top level of for next time step
    dfm_last_top.vec.data = dfm_current_top.vec
    Redraw()
print("\nL2 max: ", l2max, " Time int: ", l2_timeintmax)
print("\ndist max: ", max(maxdists))



"""
In this example we solve a scalar *unfitted* PDE problem on a moving
domain. The setting is the same as in `spacetimeDG_unfitted.py`
However, here, we apply a continuous-in-time Petrov-Galerkin method.

Domain + PDE problem:
---------------------
As in `spacetimeDG_unfitted.py`

Discretisation:
---------------
* Background space-time finite element space restricted to active domain

* Ghost penalty stabilization to deal with bad cuts (version as in [1])
  and in order to define a proper extension to a neighborhood, cf. [2]

Implementational aspects (cf. [1] and [2] for details):
-------------------------------------------------------
* Geometry approximation in space-time using isoparametric unfitted FEM

* Projection operator that maps solutions from one deformed mesh to another

* A (sparse) direct solver is applied to solve the arising linear systems.

References:
-----------
All concepts that are used here are explained in the jupyter-tuorials
`spacetime.ipynb`. As a simplified setting without cut configurations,
we also refer to the `spacetimeDG_fitted.py` demo.

Literature:
-----------
[1] J. Preuß, Higher order unfitted isoparametric space-time FEM on moving
    domains. Master's thesis, NAM, University of Göttingen, 2018.
[2] F. Heimann. On Discontinuous- and Continuous-In-Time Unfitted Space-Time
    Methods for PDEs on Moving Domains. Master's thesis, NAM, University of
    Göttingen, 2020.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from ngsolve import *
from netgen.geom2d import SplineGeometry
from xfem import *
from math import pi
from xfem.lset_spacetime import *
ngsglobals.msg_level = 1

# -------------------------------- PARAMETERS ---------------------------------
# DISCRETIZATION PARAMETERS:

# Parameter for refinement study:
i = 2
n_steps = 2**i
space_refs = i

# Polynomial order in time
k_t = 3
# Polynomial order in space
k_s = k_t
# Polynomial order in time for level set approximation
lset_order_time = k_t
# Integration order in time
time_order = 2 * k_t
# Time stepping parameters
tstart = 0
tend = 0.5
delta_t = (tend - tstart) / n_steps
maxh = 0.5
# Ghost-penalty parameter
gamma = 0.05
# Map from reference time to physical time
told = Parameter(tstart)
t = told + delta_t * tref

# PROBLEM SETUP:

# Outer domain:
rect = SplineGeometry()
rect.AddRectangle([-0.6, -1], [0.6, 1])

# Level set geometry
# Radius of disk (the geometry)
R = 0.5
# Position shift of the geometry in time
rho = (1 / (pi)) * sin(2 * pi * t)
# Convection velocity:
w = CoefficientFunction((0, rho.Diff(t)))
max_velocity = 2
# Level set
r = sqrt(x**2 + (y - rho)**2)
levelset = r - R

# Diffusion coefficient
alpha = 1
# Exact solution
u_exact = cos(pi * r / R) * sin(pi * t)
# R.h.s.
coeff_f = (u_exact.Diff(t)
           - alpha * (u_exact.Diff(x).Diff(x) + u_exact.Diff(y).Diff(y))
           + w[0] * u_exact.Diff(x) + w[1] * u_exact.Diff(y)).Compile()

# ----------------------------------- MAIN ------------------------------------
ngmesh = rect.GenerateMesh(maxh=maxh, quad_dominated=False)
for j in range(space_refs):
    ngmesh.Refine()
mesh = Mesh(ngmesh)

# Spatial FESpace for solution
fes = H1(mesh, order=k_s, dgjumps=True)
# Time finite elements (nodal!)
tfe_i = ScalarTimeFE(k_t, skip_first_node=True)  # interior shapes
tfe_e = ScalarTimeFE(k_t, only_first_node=True)  # exterior shapes (init. val.)
tfe_t = ScalarTimeFE(k_t - 1)                    # test shapes
# Space-time finite element space
st_fes_i, st_fes_e, st_fes_t = [tfe * fes for tfe in [tfe_i, tfe_e, tfe_t]]

# Space time version of Levelset Mesh Adaptation object. Also offers integrator
# helper functions that involve the correct mesh deformation
lsetadap = LevelSetMeshAdaptation_Spacetime(mesh, order_space=k_s,
                                            order_time=lset_order_time,
                                            threshold=0.5,
                                            discontinuous_qn=True)

# lset epsilon perturbation for extended facet patch bfi domain
eps = 2 * max_velocity * delta_t

gfu_i = GridFunction(st_fes_i)
gfu_e = GridFunction(st_fes_e)

u_last = CreateTimeRestrictedGF(gfu_e, 0)

u_i = st_fes_i.TrialFunction()
u_e = st_fes_e.TrialFunction()
v_t = st_fes_t.TestFunction()

scene = DrawDC(lsetadap.levelsetp1[TOP], u_last, 0, mesh, "u_last",
               deformation=lsetadap.deformation[TOP])

lset_p1_slice = GridFunction(lsetadap.levelsetp1[BOTTOM].space)

h = specialcf.mesh_size

ba_strip = BitArray(mesh.ne)
ba_plus_hasneg = BitArray(mesh.ne)
ba_minus_haspos = BitArray(mesh.ne)
ba_facets = BitArray(mesh.nfacet)
ci = CutInfo(mesh, time_order=time_order)
ci_slice = CutInfo(mesh)

dQ = delta_t * dCut(lsetadap.levelsetp1[INTERVAL], NEG, time_order=2 * k_t,
                    deformation=lsetadap.deformation[INTERVAL],
                    definedonelements=ci.GetElementsOfType(HASNEG))
dOmnew = dCut(lsetadap.levelsetp1[TOP], NEG,
              deformation=lsetadap.deformation[TOP],
              definedonelements=ci.GetElementsOfType(HASNEG), tref=1)
dw = delta_t * dFacetPatch(definedonelements=ba_facets, time_order=time_order,
                           deformation=lsetadap.deformation[INTERVAL])


def dt(u):
    return 1.0 / delta_t * dtref(u)


a_i = RestrictedBilinearForm(trialspace=st_fes_i, testspace=st_fes_t,
                             element_restriction=ci.GetElementsOfType(HASNEG),
                             facet_restriction=ba_facets,
                             check_unused=False)

a_i += v_t * (dt(u_i) - dt(lsetadap.deform) * grad(u_i)) * dQ
a_i += (alpha * InnerProduct(grad(u_i), grad(v_t))) * dQ
a_i += (v_t * InnerProduct(w, grad(u_i))) * dQ
a_i += h**(-2) * (1 + delta_t / h) * gamma * \
    (u_i - u_i.Other()) * (v_t - v_t.Other()) * dw

f = LinearForm(st_fes_t)
f += coeff_f * v_t * dQ
f += -v_t * (dt(gfu_e) - dt(lsetadap.deform) * grad(gfu_e)) * dQ
f += -(alpha * InnerProduct(grad(gfu_e), grad(v_t))) * dQ
f += -(v_t * InnerProduct(w, grad(gfu_e))) * dQ

# Set initial values
u_last.Set(fix_tref(u_exact, 0))
# Project u_last at the beginning of each time step
lsetadap.ProjectOnUpdate(u_last)

ba_plus_hasneg_old, els_test = BitArray(mesh.ne), BitArray(mesh.ne)
ba_plus_hasneg_old.Set()

while tend - told.Get() > delta_t / 2:
    lsetadap.CalcDeformation(levelset)
    gfu_e.Set(u_last)

    # Update markers in (space-time) mesh
    ci.Update(lsetadap.levelsetp1[INTERVAL], time_order=time_order)

    # Re-evaluate the "active dofs" in the space time slab
    InterpolateToP1(lsetadap.levelsetp1[BOTTOM] - eps, lset_p1_slice)
    ci_slice.Update(lset_p1_slice)
    ba_plus_hasneg[:] = ci_slice.GetElementsOfType(HASNEG)

    InterpolateToP1(lsetadap.levelsetp1[BOTTOM] + eps, lset_p1_slice)
    ci_slice.Update(lset_p1_slice)
    ba_minus_haspos[:] = ci_slice.GetElementsOfType(HASPOS)

    ba_strip[:] = ba_minus_haspos & ba_plus_hasneg
    ba_facets[:] = GetFacetsWithNeighborTypes(mesh, a=ba_strip,
                                              b=ba_plus_hasneg)
    active_dofs = GetDofsOfElements(st_fes_i, ba_plus_hasneg)

    # Check element history for method of lines time-derivative approx.
    els_test[:] = ci.GetElementsOfType(HASNEG) & ~ba_plus_hasneg_old
    assert sum(els_test) == 0, 'Some active elements do not have a history.\
        You might want to increase eps'

    ba_plus_hasneg_old[:] = ba_plus_hasneg

    a_i.Assemble(reallocate=True)
    f.Assemble()

    # Solve linear system
    gfu_i.vec.data = a_i.mat.Inverse(active_dofs) * f.vec

    # Evaluate upper trace of solution for
    #  * for error evaluation
    #  * upwind-coupling to next time slab
    RestrictGFInTime(spacetime_gf=gfu_i, reference_time=1.0, space_gf=u_last)

    # Eompute error at final time
    l2error = sqrt(
        Integrate((u_exact - u_last)**2 * dOmnew, mesh))

    # Update time variable (ParameterCF)
    told.Set(told.Get() + delta_t)
    print("\rt = {0:12.9f}, L2 error = {1:12.9e}".format(told.Get(), l2error))

    try:
        __builtin__
        __IPYTHON__
        scene.Redraw()
    except NameError:
        scene.Redraw(blocking=True)



from ngsolve import *
from netgen.geom2d import SplineGeometry

from xfem import *
from math import pi

from xfem.lset_spacetime import *

ngsglobals.msg_level = 1

square = SplineGeometry()
square.AddRectangle([-1,-1],[1,1])
ngmesh = square.GenerateMesh(maxh=0.02, quad_dominated=False)
mesh = Mesh (ngmesh)

coef_told = Parameter(0)
coef_delta_t = Parameter(0)
tref = ReferenceTimeVariable()
t = coef_told + coef_delta_t*tref

r0 = 0.5
r = sqrt(x**2+y**2+t**2)

# level set
levelset= r - r0

time_order = 1
fes1 = H1(mesh, order=1)
tfe = ScalarTimeFE(time_order)
st_fes = SpaceTimeFESpace(fes1,tfe)

n_steps = 100

tend = 1
delta_t = tend/n_steps
coef_delta_t.Set(delta_t)
told = 0

lset_p1 = GridFunction(st_fes)

sum_vol = 0
sum_int = 0
for i in range(n_steps):
    #SpaceTimeInterpolateToP1(levelset,coef_told,told,delta_t,lset_p1) #call for the master Branch
    SpaceTimeInterpolateToP1(levelset,tref,0.,delta_t,lset_p1) # call for the master spacetime_weihack -- 0 and tend are ununsed parameter
    
    val_vol = Integrate({ "levelset" : lset_p1, "domain_type" : NEG}, CoefficientFunction(1.0), mesh, time_order = time_order)
    val_int = Integrate({ "levelset" : lset_p1, "domain_type" : IF}, CoefficientFunction(1.0), mesh, time_order = time_order)
    #print(val_vol, val_int)
    sum_vol += val_vol*delta_t
    sum_int += val_int*delta_t
    
    told = told + delta_t
    coef_told.Set(told)

print("SUM VOL: ", sum_vol)
print("VOL: ", 2/3*pi*r0**3)
print("\t\tDIFF: ", abs(sum_vol - 2/3*pi*r0**3))

print("SUM INT: ", sum_int)
print("AREA: ", 0.5*pi**2*r0**2)
print("\t\tDIFF: ", abs(sum_int - 0.5*pi**2*r0**2))



from test_spacetimecutrule import test_spacetime_spaceP4_timeDGP4
from xfem import ngsxfemglobals

def test_multiply_all_eps():
    e1 = test_spacetime_spaceP4_timeDGP4()

    ngsxfemglobals.MultiplyAllEps(1e4)

    e2 = test_spacetime_spaceP4_timeDGP4()

    assert( abs(e1-e2) > 1e-10)



"""
In this example we integrate over a circle with radius 1 to illustrate
the higher order convergence achieved through the isoparametric mapping
approach. The mesh is deformed in order to map the P1 level set
approximation onto a higher-order approximation of the smooth level set,
thereby overcoming the geometry error of h^2 made by the piecewise
linear level set approximation and recovering the full integration
order.

Literature
----------
[1] C. Lehrenfeld. High order unfitted finite element methods on level
    set domains using isoparametric mappings. Comp. Meth. Appl. Mech.
    Eng., 300:716-733, 2016.
[2] C. Lehrenfeld. A higher order isoparametric fictitious domain method
    for level set domains. In S. P. A. Bordas, E. Burman, M. G. Larson,
    and M. A. Olshanskii, editors, Geometrically Unfitted Finite Element
    Methods and Applications, pages 65-92. Springer International
    Publishing, 2017.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *

from xfem import *
from xfem.lsetcurv import *

from math import pi


# -------------------------------- PARAMETERS ---------------------------------
maxh = 0.5
order = 5

maxreflvl = 6


# ----------------------------------- MAIN ------------------------------------
def Make2DProblem(maxh=2):
    square = SplineGeometry()
    square.AddRectangle([-1, -1], [1, 1], bc=1)
    mesh = Mesh(square.GenerateMesh(maxh=maxh, quad_dominated=False))
    return mesh


# Circle with radius 0.5
levelset = sqrt(x * x + y * y) - 0.5
integrand = CoefficientFunction(1.0)
referencevals = {POS: 4.0 - 0.25 * pi, NEG: 0.25 * pi, IF: pi}

mesh = Make2DProblem(maxh=maxh)

# Level set mesh adaptation machinery
lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.2,
                                      discontinuous_qn=True)
lsetp1 = lsetmeshadap.lset_p1

# Containers to compute error convergence
err_uncurved, err_curved = {}, {}
eoc_uncurved, eoc_curved = {}, {}

for key in [NEG, POS, IF]:
    err_curved[key] = []
    err_uncurved[key] = []
    eoc_curved[key] = []
    eoc_uncurved[key] = []


# Main loop
for reflevel in range(maxreflvl):
    if(reflevel > 0):
        mesh.Refine()
    lsetmeshadap.CalcDeformation(levelset)

    for key in [NEG, POS, IF]:
        dx = dCut(lsetp1, key, order=order)
        dy = dCut(lsetp1, key, order=order, deformation=lsetmeshadap.deform)
        integral_uncurved = Integrate(integrand*dx, mesh)
        integral_curved = Integrate(integrand*dy, mesh)

        err_curved[key].append(abs(integral_curved - referencevals[key]))
        err_uncurved[key].append(abs(integral_uncurved - referencevals[key]))

    # Mark cut elements for refinement:
    RefineAtLevelSet(gf=lsetmeshadap.lset_p1)

for key in [NEG, POS, IF]:
    eoc_curved[key] = [log(a / b) / log(2) for (a, b)
                       in zip(err_curved[key][:-1], err_curved[key][1:])]
    eoc_uncurved[key] = [log(a / b) / log(2) for (a, b)
                         in zip(err_uncurved[key][:-1], err_uncurved[key][1:])]

print("errors (uncurved):  \n{}\n".format(err_uncurved))
print("   eoc (uncurved):  \n{}\n".format(eoc_uncurved))
print("errors (  curved):  \n{}\n".format(err_curved))
print("   eoc (  curved):  \n{}\n".format(eoc_curved))

Draw(levelset, mesh, "levelset")
Draw(lsetmeshadap.deform, mesh, "deformation")
Draw(lsetmeshadap.lset_p1, mesh, "levelset(P1)")



"""
In this example we solve a scalar *unfitted* PDE problem as in fictdom.py
but with an unfitted DG discretization.

Domain + PDE problem:
---------------------
As in fictdom.py

Discretisation:
---------------
* Background discontinuous FEspace restricted to active domain (CutFEM)

* Interior penalty DG formulation inside the domain, see e.g. [1]

* Nitsche formulation to impose boundary conditions, see. e.g. [2]

* Ghost penalty stabilization to deal with bad cuts (the so-called
  "direct" version as in [3])

Implementational aspects:
-------------------------
* Geometry approximation using isoparametric unfitted FEM

* A (sparse) direct solver is applied to solve the arising linear systems.

References:
-----------
* All concepts that are used here are explained in the jupyter tutorials
  `basics`, `intlset` and `cutfem`.


Literature:
-----------
[1] C. Gürkan, S. Sticko, A. Massing, Stabilized Cut Discontinuous Galerkin
    Methods for Advection-Reaction Problems, SIAM J. Sci. Comp. 42(3):
    A2620-A2654, 2020.
[2] E. Burman, P. Hansbo, Fictitious domain finite element methods using
    cut elements: II. A stabilized Nitsche method, Appl. Num. Math.
    62(4):328-341, 2012.
[3] J. Preuß, Higher order unfitted isoparametric space-time FEM on
    moving domains. Master's thesis, NAM, University of Göttingen, 2018.
"""

# ------------------------------ LOAD LIBRARIES -------------------------------
from netgen.geom2d import SplineGeometry
from ngsolve import *
from ngsolve.internal import *
from xfem import *
from xfem.lsetcurv import *

ngsglobals.msg_level = 2

# -------------------------------- PARAMETERS ---------------------------------
# Quadrilateral (or simplicial mesh)
quad_mesh = False
# Mesh diameter
maxh = 0.1
# Finite element space order
order = 2

# Stabilization parameter for ghost-penalty
gamma_stab = 0.1
# Stabilization parameter for Nitsche
lambda_nitsche = 10 * order * order
# Stabilization parameter for interior penalty
lambda_dg = 10 * order * order

# ----------------------------------- MAIN ------------------------------------
# Geometry and Mesh
square = SplineGeometry()
square.AddRectangle((-1, -1), (1, 1), bc=1)

ngmesh = square.GenerateMesh(maxh=maxh, quad_dominated=quad_mesh)
mesh = Mesh(ngmesh)


# Manufactured exact solution for monitoring the error
r2 = 3 / 4  # outer radius
r1 = 1 / 4  # inner radius
rc = (r1 + r2) / 2.0
rr = (r2 - r1) / 2.0
r = sqrt(x**2 + y**2)
levelset = IfPos(r - rc, r - rc - rr, rc - r - rr)

exact = (20 * (r2 - sqrt(x**2 + y**2)) * (sqrt(x**2 + y**2) - r1)).Compile()
coeff_f = - (exact.Diff(x).Diff(x) + exact.Diff(y).Diff(y)).Compile()

# Higher order level set approximation
lsetmeshadap = LevelSetMeshAdaptation(mesh, order=order, threshold=0.1,
                                      discontinuous_qn=True)
deformation = lsetmeshadap.CalcDeformation(levelset)
lsetp1 = lsetmeshadap.lset_p1

# Element, facet and dof marking w.r.t. boundary approximation with lsetp1:
ci = CutInfo(mesh, lsetp1)
hasneg = ci.GetElementsOfType(HASNEG)
hasif = ci.GetElementsOfType(IF)
# Facets for ghost penalty stabilization
ba_gp_facets = GetFacetsWithNeighborTypes(mesh, a=hasneg, b=hasif)
# Facets with parts inside the domain
ba_fd_facets = GetFacetsWithNeighborTypes(mesh, a=hasneg, b=hasneg)

Vhbase = L2(mesh, order=order, dirichlet=[], dgjumps=True)
Vh = Restrict(Vhbase, hasneg)

gfu = GridFunction(Vh)


u, v = Vh.TnT()
h = specialcf.mesh_size
nh = 1.0 / Norm(grad(lsetp1)) * grad(lsetp1)
nF = specialcf.normal(mesh.dim)
flux_u = -0.5 * (grad(u) + grad(u.Other())) * nF
flux_v = -0.5 * (grad(v) + grad(v.Other())) * nF
jump_u = u - u.Other()
jump_v = v - v.Other()


# Element-wise integrals
dx = dCut(lsetp1, NEG, definedonelements=hasneg, deformation=deformation)
# Interior skeleton integrals:
dk = dCut(lsetp1, NEG, skeleton=True, definedonelements=ba_fd_facets,
          deformation=deformation)
# Domain boundary integrals
ds = dCut(lsetp1, IF, definedonelements=hasif, deformation=deformation)
# Ghost penalty integrals
dw = dFacetPatch(definedonelements=ba_gp_facets, deformation=deformation)

a = RestrictedBilinearForm(Vh, "a", hasneg, ba_fd_facets, check_unused=False)
a += grad(u) * grad(v) * dx
a += (lambda_dg / h * jump_u * jump_v + flux_u * jump_v + flux_v * jump_u) * dk
a += (-grad(u) * nh * v - grad(v) * nh * u + lambda_nitsche / h * u * v) * ds
a += gamma_stab / h**2 * (u - u.Other()) * (v - v.Other()) * dw

f = LinearForm(Vh)
f += coeff_f * v * dx

# Assemble system
f.Assemble()
a.Assemble()

# Solve linear system
gfu.vec.data = a.mat.Inverse(Vh.FreeDofs(), "sparsecholesky") * f.vec

# Measure the error
l2error = sqrt(Integrate((gfu - exact)**2 * dx, mesh))
print("L2 Error: {0}".format(l2error))

# Unset mesh adaptation
mesh.deformation = None

# Visualization:
Draw(deformation, mesh, "deformation")
DrawDC(lsetp1, gfu, 0, mesh, "uh")
